/*
 *         _        ___
 *    /\ / /___    / _ \ /\ /\  _ ___
 *   / // // _ \  / // // // // // _ \
 *  / // // // / / ___// // // // // /
 *  \___//_//_/ /_/   /_/ \___/ \_  /
 *                             \___/
 * 
 *  Compunach UnPlug
 *  Copyright (C) 2009 David Batley <unplug@dbatley.com>
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 2 of the License, or
 *  (at your option) any later version.
 * 
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 * 
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */


UnPlug2Rules = {
	/*
	 * Executes an if_* statement.
	 * statement_name should not begin with "if_"
	 * If there isn't a suitable if_* statement, one will be made by returning the result from appropriate each_* functions
	 */
	exec_if : function (statement_name, node, variables, url, text, doc) {
		var rule = UnPlug2Rules.get_rule("if_" + statement_name);
		
		// if no <if...>, try and return the first part of the <each...> statement.
		if (!rule) {
			// response list is already prefixed variables with ref
			var response_list = UnPlug2Rules.exec_each(statement_name, node, variables, url, text, doc);
			if (response_list.length < 1)
				throw statement_name + " returned empty list"
			return response_list[0];
		}
		
		// exec the <if...>
		var response = UnPlug2Rules.exec_statement(rule, node, variables, url, text, doc);
		
		// valid responses: true, false, object_of_new_variables_without_prefixes
		if (response == false)
			// throw statement_name + " returned false";
			return false;
		if (response == true)
			return {};
		if (rule.apply_ref) {
			var ref = node.getAttribute("ref");
			return UnPlug2Rules._prefix_variables(ref, response);
		}
		// without apply_ref set this responds by exporting variables without the prefix!
		// TODO - is this right behaviour?
		return response;
	},
	
	/*
	 * Executes an optional_* statement.
	 * statement_name should not begin with "optional_"
	 * This is actually just a wrapper for if_* statements.
	 */
	exec_optional : function (statement_name, node, variables, url, text, doc) {
		try {
			return UnPlug2Rules.exec_if(statement_name, node, variables, url, text, doc);
		} catch (e) {
			// it's optional, we don't care about failure.
		}
	},
	
	/*
	 * Executes an each_* statement.
	 * statement_name should not begin with "each_"
	 */
	exec_each : function (statement_name, node, variables, url, text, doc) {
		var rule = UnPlug2Rules.get_rule("each_" + statement_name);
		var response_list = UnPlug2Rules.exec_statement(rule, node, variables, url, text, doc);
		var ref = node.getAttribute("ref");
		for (var i=0; i < response_list.length; i++) {
			if (!rule.apply_ref) {
				// without apply_ref set this responds by exporting variables without the prefix!
				// TODO - is this right behaviour?
				// pass -- just leave as is
			} else  {
				response_list[i] = UnPlug2Rules._prefix_variables(ref, response_list[i]);
			}
		}
		return response_list;
	},
	
	/*
	 * Return an object describing the rule
	 */
	get_rule : function (rulename) {
		return UnPlug2Rules[rulename];
	},
	
	/*
	 * Private.
	 * puts "ref" in front of results
	 */
	_prefix_variables : function (ref, response) {
		if (!ref)
			return {};
		var response_with_ref = {};
		for (k in response) {
			response_with_ref[ref + "." + k] = response[k];
		}
		return response_with_ref;
	},
	
	/**
	 * Execute a statement.
	 * 
	 * IMPORTANT: You should not normally call this function - instead you should call the exec_if(), exec_optional() or exec_each() functions as appropriate.
	 * 
	 * This will return an object containing new variables which are not prefixed by $ref, or throw on failure.
	 * Rule is one of the rules, which you can get with get_rule().
	 * node is the dom element of the statement from rules.xml (so we can get attributes)
	 * variables is the variables available to this scope
	 */
	exec_statement : function (rule, node, variables, url, text, doc) {
		if (!rule)
			throw statement_name + " is not a rule which has been implemented in " + UnPlug2.version + " " + UnPlug2.revision;
		// use rule.required and rule.optional to make a data object to feed to rule.execute
		var data = {}
		if (rule.required) {
			for (var i=0; i < rule.required.length; i++) {
				var attrname = rule.required[i];
				var value = null;
				if (attrname == "innerHTML") {
					value = UnPlug2.trim(node.textContent);
				} else {
					value = node.getAttribute(attrname);
				}
				if (!value)
					throw "Attribute " + attrname + " required";
				data[attrname] = variables.subst(value);
			}
		}
		if (rule.optional) {
			for (var i=0; i < rule.optional.length; i++) {
				var attrname = rule.optional[i];
				var value = null;
				if (attrname == "innerHTML") {
					value = node.textContent;
				} else {
					value = node.getAttribute(attrname);
				}
				try {
					data[attrname] = variables.subst(value ? value : "");
				} catch (e) {
					// problems substing variables is fine as it's optional
					// eg <media thumbnail="..."> - errors substing thumbnail won't take down rule.
					data[attrname] = "";
				}
			}
		}
		// exec function - may throw, return true or false, or return an object
		return rule.execute(data, url, text, doc);
	},
	
	if_version : {
		order     : 10,
		optional  : ["min", "max"],
		execute   : function (data, url, text, doc) {
			if (data.min && UnPlug2.version < parseFloat(data.min))
				return false;
			if (data.max && UnPlug2.version >= parseFloat(data.max))
				return false;
			return true;
		}
	},
	
	if_revision : {
		// order decides which rule is done first. if_ is done 0..100; download/media/etc 500...600
		order     : 10,
		optional  : ["min", "max"],
		execute   : function (data, url, text, doc) {
			if (UnPlug2.revision == null)
				return true;
			if (data.min && UnPlug2.revision < parseFloat(data.min))
				return false;
			if (data.max && UnPlug2.revision >= parseFloat(data.max))
				return false;
			return true;
		}
	},
	
	/*
	 * Private.
	 * Takes the response to a RegExp.exec (or String.match) and creates an object containing {"1" : "subexpression 1 result", "2" : ... };
	 */
	_re_resp : function (re_resp) {
		if (!re_resp)
			return false; // no matches
		var result = {};
		for (var i=1; i < re_resp.length; i++)
			if (re_resp[i])
				result[i] = re_resp[i];			
		return result;
	},
	
	/*
	 * Matches a regular expression contained in innerHTML
	 */
	if_re : {
		order     : 90,
		required  : ["innerHTML"],
		optional  : ["string", "flags", "tagname"],
		apply_ref : true,
		execute   : function (data, url, text, doc) {
			var re = RegExp(data.innerHTML, data.flags);
			
			if (data.tagname) {
				if (!doc)
					throw "No document for if_re with elemname";
				for each (elem in doc.getElementsByTagName(data.tagname)) {
					var r = UnPlug2Rules._re_resp(re.exec(elem.textContent));
					if (r)
						return r;
				}
				return false;
			} else {
				return UnPlug2Rules._re_resp(re.exec(data.string || text))
			}
		}
	},
	
	each_re : {
		order     : 90,
		required  : ["re"],
		optional  : ["string", "flags", "tagname"],
		apply_ref : true,
		execute   : function (data, url, text, doc) {
			var re = RegExp(data.re, data.flags + "g");
			var response = [];
			var texts = [];
			if (data.tagname) {
				if (!doc)
					throw "No document for if_re with elemname";
				for each (elem in elemlist) {
					texts[texts.length] = elem.textContent;
				}
			} else {
				texts = [ data.string || text ];
			}
			for (var i = 0; i < texts.length; i++) {
				var rr;
				while (rr = re.exec(texts[i])) {
					response[response.length] = UnPlug2Rules._re_resp(rr);
				}
			}
			return response;
		}
	},

	/*
	 * Checks the url.
	 */
	if_url : {
		order     : 20,
		optional  : [
			"inhost", // substring is in host
			"inurl",  // substring is in spec
			"inpath", // substring is in path
			"path",   // matches start of path
			"port",   // matches port exactly
			"host"],   // matches end of host in whole words
		apply_ref : true, // populate $ref = url.spec, $ref.host = url.host, etc....
		execute   : function (data, url, text, doc) {
			if (data.inhost) {
				if (url.host.indexOf(data.inhost) < 0)
					return false;
			}
			if (data.inurl) {
				if (url.spec.indexOf(data.inurl) < 0)
					return false;
			}
			if (data.inpath) {
				if (url.path.indexOf(data.inpath) < 0)
					return false;
			}
			if (data.port) {
				if (url.port != data.port)
					return false;
			}
			if (data.protocol) {
				if (url.scheme != data.protocol)
					return false;
			}
			if (data.path) {
				if (url.path.indexOf(data.path) != 0)
					return false;
			}
			if (data.host) {
				/*
				 * TODO - this isn't implemented properly
				 * examle of host
				 * eg: foo.com
				 * matches foo.com
				 * matches www.yay.foo.com
				 * doesn't match foo.com.nyud.net
				 * doesn't match xyay.com
				 * possibly matches yay.com.
				 */
				if (url.host.indexOf(data.host) < 0)
					return false;
			}
			return {
				"url"      : url.spec,
				"host"     : url.host,
				"post"     : url.port,
				"path"     : url.path,
				"protocol" : url.scheme }
		}
	},
	
	/*
	 * Iterates over the dom elements in doc which match the specified pattern
	 */
	each_element : {
		order     : 70,
		optional  : [
			"tagname",
			"elemid",
			"require_attrs",
			"attrs"],
		apply_ref : true,
		execute   : function (data, url, text, doc) {
			if (!doc)
				throw "each_element requires a document";
			
			// apply elemid and tagname
			if (data.elemid) {
				var elem = doc.getElementById(data.elemid);
				if (!elem)
					return false;
				if (data.tagname && data.tagname != elem.tagName)
					throw "Element with id did not match tagname";
				var elem_list = [elem];
			} else if (data.tagname) {
				var elem_list = doc.getElementsByTagName(data.tagname);
			} else {
				throw "each_element has no tagname and no elemid";
			}
			
			// elem_list is populated with a list of possible elements
			var result_list = [];
			for (var i=0; i < elem_list.length; i++) {
				var elem = elem_list[i];
				
				// get the variables that are in response of this one element
				var result = {};
				var got_all_required = true;
				for each (req_attr in data.require_attrs.split(",")) {
					switch (req_attr) {
						case "innerHTML" : 
							result.innerHTML = elem.textContent;
							break;
						case "":
							break;
						default:
							var attrvalue = elem.getAttribute(req_attr);
							if (attrvalue)
								result[req_attr] = attrvalue;
							else
								got_all_required = false;
							break;
					}
				}
				if (!got_all_required) {
					continue;
				}
				for each (attr in data.attrs.split(",")) {
					switch (attr) {
						case "innerHTML" : 
							result.innerHTML = elem.textContent;
							break;
						case "":
							break;
						default:
							var attrvalue = elem.getAttribute(attr);
							if (attrvalue)
								result[attr] = attrvalue;
							break;
					}
				}
				
				// add this one element's response to the list
				result_list[result_list.length] = result;
			}
			
			return result_list;
		}
	},
	
	_media_ext : [
		"wma", "wmv",
		"mpg", "mpeg", "mp3", "mp2", "mp4",
		"divx", "avi",
		"ogg", "ogv", "flac", "speex",
		"mkv",
		"flv" ],
	
	if_ext : {
		order     : 70,
		required  : ["filename"],
		optional  : [
			"ext",
			"is_media"],
		apply_ref : true,
		execute   : function (data, url, text, doc) {
			var fullname = data.filename.toLowerCase();
			
			// strip gubbinns from end of url
			var lasthash = fullname.lastIndexOf("#");
			if (lasthash > 0)
				fullname = fullname.substring(0, lasthash);
			var lastquestion = fullname.lastIndexOf("?");
			if (lastquestion > 0)
				fullname = fullname.substring(0, lastquestion);
			var lastsemi = fullname.lastIndexOf(";");
			if (lastsemi > 0)
				fullname = fullname.substring(0, lastsemi);
		
			// get ext
			var lastdot = fullname.lastIndexOf(".");
			if (lastdot < 0)
				return false;
		
			var ext = fullname.substring(lastdot+1).toLowerCase();
			if (ext.length < 1 || ext.length > 5)
				return false;
		
			if (data.is_media) {
				if (UnPlug2Rules._media_ext.indexOf(ext) < 0)
					return false;
			}
		
			// return true
			return { ref : ext };
		}
	},
	
	end_of_object : 1 };
